#!/usr/bin/python
import time, struct, sys, os, getopt
from ConfigParser import ConfigParser
from bsddb3 import db as bdb
from socket import inet_ntoa
from IPy import IP

class IPTree(IP):
	def __init__(self, net = "0/0", dst = "0/0"):
		IP.__init__(self, net)
		self.dst = IP(dst)
		self.nodes = []

	def add(self, entry, dst):
		entry = IP(entry)
		dst = IP(dst)
		if self.prefixlen() == entry.prefixlen():
			return
		begin = entry.net()
		end = entry.broadcast()
		for i in xrange(len(self.nodes)):
			if begin > self.nodes[i].broadcast():
				continue
			if end < self.nodes[i].net():
				self.nodes.insert(i, IPTree(entry, dst))
				return
			if entry.prefixlen() >= self.nodes[i].prefixlen():
				return self.nodes[i].add(entry, dst)
			node = IPTree(entry, dst)
			while i < len(self.nodes) and end >= self.nodes[i].net():
				node.nodes.append(self.nodes.pop(i))
			self.nodes.insert(i, node)
			return
		self.nodes.append(IPTree(entry, dst))

	def lookup(self, entry):
		ip = IP(entry).net()
		for node in self.nodes:
			if ip > node.broadcast():
				continue
			if ip < node.net():
				break
			return node.lookup(ip)
		delta = ip.int() - self.int()
		delta = delta % self.dst.len()
		return IP(self.dst.int() + delta)

	def dump(self, pfx = ''):
		print "%s%s -> %s" % (pfx, self, self.dst)
		for i in self.nodes:
			i.dump(pfx + '  ')

def log(msg):
	print msg

class binary:
	def __init__(self, str):
		self.str = str
		self.pos = 0
		self.len = len(str)

	def getInt(self, size):
		self.pos = (self.pos + size - 1) & ~(size - 1)
		if self.pos + size > self.len:
			raise Exception("short read")
		val = 0
		for i in range(size, 0, -1):
			val = (val << 8) | ord(self.str[i + self.pos - 1])
		self.pos = self.pos + size
		return val

	def getString(self, size):
		if self.pos + size > self.len:
			raise Exception("short read")
		str = self.str[self.pos:self.pos + size]
		self.pos = self.pos + size
		pos = str.find("\0")
		if str >= 0:
			return str[:pos]
		else:
			return str

class entry:
	def __init__(self, db):
		self.cursor = db.cursor()

	def __del__(self):
		self.cursor.close()

	def parse(str):
		str = binary(str)
		version = str.getInt(4)
		if version != 1:
			raise Exception("Unsupported entry data version %d" % version)
		res = {}
		res['oob.time.sec'] = str.getInt(4)
		res['oob.time.usec'] = str.getInt(4)
		res['ip.saddr'] = str.getInt(4)
		res['ip.daddr'] = str.getInt(4)
		res['ip.protocol'] = str.getInt(1)
		res['tcp.flags'] = str.getInt(1)
		res['ip.totlen'] = str.getInt(2)
		res['tcp.sport'] = str.getInt(2)
		res['tcp.dport'] = str.getInt(2)
		res['udp.sport'] = str.getInt(2)
		res['udp.dport'] = str.getInt(2)
		res['udp.len'] = str.getInt(2)
		res['icmp.type'] = str.getInt(1)
		res['icmp.code'] = str.getInt(1)
		res['icmp.echoid'] = str.getInt(2)
		res['icmp.echoseq'] = str.getInt(2)
		res['icmp.gateway'] = str.getInt(4)
		res['icmp.fragmtu'] = str.getInt(2)
		res['raw.mac'] = str.getString(80)
		return res
	parse = staticmethod(parse)

	def next(self):
		global delEntries
		data = self.cursor.next()
		if data == None:
			raise StopIteration
		if delEntries:
			self.cursor.delete()
		try:
			return entry.parse(data[1])
		except Exception, e:
			log(e)
			return None

class dbfile:
	def __init__(self, envdir, db):
		self.env = bdb.DBEnv()
		self.env.open(envdir, bdb.DB_CREATE | bdb.DB_INIT_LOCK | bdb.DB_INIT_MPOOL)

		self.db = bdb.DB(self.env)
		self.db.set_flags(bdb.DB_DUP)
		self.db.open(db, dbtype = bdb.DB_BTREE)

	def __del__(self):
		self.db.close()
		self.env.close()

	def __iter__(self):
		return entry(self.db)

def intToIp(addr):
	return inet_ntoa(struct.pack('>L', addr))

def getTz():
	tz = time.timezone
	return "%c%02d:%02d" % ("+-"[tz > 0], abs(tz) / 3060, (abs(tz) / 60) % 60)

def mkline(e, tree):
	if e['ip.protocol'] == 1:
		sport = e['icmp.type']
		dport = e['icmp.code']
		flags = ''
	elif e['ip.protocol'] == 6:
		sport = e['tcp.sport']
		dport = e['tcp.dport']
		def mapFlag(idx, name):
			if e['tcp.flags'] & (1 << idx):
				return "\t" + name
			else:
				return ''
		flags = reduce(lambda a, b: a + b, \
		               map(lambda tuple: mapFlag(*tuple), \
				[(1, 'SYN'), (4, 'ACK'), (0, 'FIN'), \
				 (5, 'URG'), (2, 'RES'), (3, 'PSH')]))
	elif e['ip.protocol'] == 17:
		sport = e['udp.sport']
		dport = e['udp.dport']
		flags = ''
	else:
		sport = 0
		dport = 0
		flags = ''

	src = inet_ntoa(struct.pack('>L', e['ip.saddr']))
	dst = tree.lookup(e['ip.daddr'])
	if not dst:
		return None

	return "%s\t%d\t%s\t%d\t%d%s" % \
		(src, sport, dst, dport, e['ip.protocol'], flags)

def header(fromMail, toMail, id, tz):
	print "From: %s" % fromMail
	print "To: %s" % toMail
	print "Subject: FORMAT DSHIELD USERID %s TZ %s" % (id, tz)
	print "Content-Type: text/plain; charset=ascii"
	print "Content-Transfer-Encoding: 7bit"
	print ""

def main():
	global delEntries
	delEntries = False

	optlist, args = getopt.getopt(sys.argv[1:], 'dh', ['delete', 'help'])
	for o, a in optlist:
		if (o == '-d' or o == '--delete'):
			delEntries = True
		elif (o == '-h' or o == '--help'):
			print 'Usage: ' + sys.argv[0] + '[OPTIONS]'
			print 'Options:'
			print '\t-d, --delete\tdelete entries from db'
			sys.exit()

	config = ConfigParser()
	config.read(['/etc/dshield-bdb.conf',			\
	             os.environ['HOME'] + '/dshield-bdb.conf',	\
	             'dshield-bdb.conf'])
	id = config.get('main', 'userid')
	fromMail = config.get('main', 'email')
	toMail = config.get('main', 'mailto')
	env = config.get('main', 'home')
	db = config.get('main', 'db')

	tree = IPTree()
	for entry in config.items('maps'):
		tree.add(*entry)

	db = dbfile(env, db)
	tz = getTz()

	lastTime = -1
	lastLine = ''
	count = 0
	headerSent = False
	for e in db:
		if e == None:
			continue
		now = e['oob.time.sec']
		line = mkline(e, tree)
		if line == None:
			continue
		if now == lastTime and line == lastLine:
			count = count + 1
			continue

		if count:
			if not headerSent:
				headerSent = True
				header(fromMail, toMail, id, tz)
			stamp = time.localtime(lastTime)
			stamp = time.strftime("%Y-%m-%d %H:%M:%S", stamp)
			print "%s %s\t%s\t%d\t%s" % \
					(stamp, tz, id, count, lastLine)

		count = 1
		lastTime = now
		lastLine = line

	if count:
		if not headerSent:
			headerSent = True
			header(fromMail, toMail, id, tz)
		stamp = time.localtime(lastTime)
		stamp = time.strftime("%Y-%m-%d %H:%M:%S", stamp)
		print "%s %s\t%s\t%d\t%s" % (stamp, tz, id, count, lastLine)

if __name__ == '__main__':
	main()
